from collections.abc import Iterable
from contextlib import AsyncContextDecorator, ContextDecorator
from contextvars import ContextVar
from dataclasses import dataclass, replace
from typing import Any, AsyncContextManager, ContextManager, NamedTuple, ParamSpec, TypeVar

from attr import frozen
from jinja2 import Template

from .function import _RESPONSE_TOOL_NAME, AIFunction
from .models import AssistantMessage, Message, Model, ToolMessage, parse_messages
from .models.model import FinishReason, GenerationParams, ResponseOption, ToolParams, ToolUseMode
from .schema import SchemaValidationError

__all__ = ["Runtime"]


_P = ParamSpec("_P")
_R = TypeVar("_R")


class WangaRuntimeError(Exception):
    pass


class ContentFilterError(ValueError, WangaRuntimeError):
    pass


class OutputTooLongError(RuntimeError, WangaRuntimeError):
    pass


def render_prompt(template: Template, arguments: dict[str, Any]) -> list[Message]:
    message_str = template.render(arguments)
    return parse_messages(message_str)


class ToolInvocationResults(NamedTuple):
    messages: list[ToolMessage]
    errors: list[Exception]


@frozen
class FinalResponse:
    value: Any


def invoke_tools(message: AssistantMessage, tool_params: ToolParams) -> ToolInvocationResults | FinalResponse:
    messages = []
    errors = []
    for invocation in message.tool_invocations:
        tool = tool_params.get_tool(invocation.name)
        try:
            tool_output = tool.eval(invocation.arguments)
            if tool.name == _RESPONSE_TOOL_NAME:
                return FinalResponse(tool_output)
        except SchemaValidationError as e:
            tool_output = str(e)
            errors.append(e)
        messages.append(ToolMessage(invocation.invocation_id, str(tool_output)))
    return ToolInvocationResults(messages, errors)


@dataclass
class InvocationState:
    messages: list[Message]
    tool_params: ToolParams
    generation_params: GenerationParams
    allow_plain_text_response: bool
    max_retries_on_invalid_output: int
    result: Any | None = None


def _handle_model_response(response: ResponseOption, state: InvocationState) -> InvocationState:
    message = response.message

    new_messages = list(state.messages)
    new_messages.append(message)
    state = replace(state, messages=new_messages)

    if response.finish_reason == FinishReason.STOP:
        if state.allow_plain_text_response:
            return replace(state, result=message.content)
    if response.finish_reason in [FinishReason.STOP, FinishReason.TOOL_CALL]:
        if message.tool_invocations:
            tool_invocation_results = invoke_tools(message, state.tool_params)
            if isinstance(tool_invocation_results, FinalResponse):
                return replace(state, result=tool_invocation_results.value)
            tool_messages, errors = tool_invocation_results
            new_messages.extend(tool_messages)
            state = replace(state, messages=new_messages)
            if errors:
                if state.max_retries_on_invalid_output == 0:
                    raise errors[0]
                state = replace(state, max_retries_on_invalid_output=state.max_retries_on_invalid_output - 1)
        else:
            raise RuntimeError("Model returned a stop response without any tool invocations.")
    elif response.finish_reason == FinishReason.CONTENT_FILTER:
        raise ContentFilterError(f"Content filter triggered: {message.content}")
    elif FinishReason.LENGTH:
        raise OutputTooLongError(message.content)

    return state


def call_and_use_tools(
    model: Model,
    messages: list[Message],
    tool_params: ToolParams,
    generation_params: GenerationParams,
    allow_plain_text_response: bool,
    max_retries_on_invalid_output: int,
) -> Any:
    state = InvocationState(
        messages,
        tool_params,
        generation_params,
        allow_plain_text_response,
        max_retries_on_invalid_output,
    )
    while True:
        response = model.reply(state.messages, state.tool_params, state.generation_params).response_options[0]
        state = _handle_model_response(response, state)
        if state.result is not None:
            return state.result


async def call_and_use_tools_async(
    model: Model,
    messages: list[Message],
    tool_params: ToolParams,
    generation_params: GenerationParams,
    allow_plain_text_response: bool,
    max_retries_on_invalid_output: int,
) -> Any:
    state = InvocationState(
        messages,
        tool_params,
        generation_params,
        allow_plain_text_response,
        max_retries_on_invalid_output,
    )
    while True:
        response = (
            await model.reply_async(state.messages, state.tool_params, state.generation_params)
        ).response_options[0]
        state = _handle_model_response(response, state)
        if state.result is not None:
            return state.result


class Runtime(ContextDecorator, AsyncContextDecorator, ContextManager, AsyncContextManager):
    def _get_model(self, model: str | Model) -> Model:
        if isinstance(model, Model):
            return model
        else:
            raise NotImplementedError

    def __init__(self, model: str | Model | Iterable[str | Model]):
        if not isinstance(model, Iterable):
            model = [model]
        self.models = [self._get_model(m) for m in model]

    async def __aenter__(self):
        self._token = _global_runtime.set(self)
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        _global_runtime.reset(self._token)

    def __enter__(self):
        self._token = _global_runtime.set(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        _global_runtime.reset(self._token)

    def _select_model(self, preferred_models: list[str]) -> Model:
        name_to_model = {model.name: model for model in self.models}
        for preferred_model in preferred_models:
            model = name_to_model.get(preferred_model)
            if model is not None:
                return model
        return self.models[0]

    def _prepare_kwargs(self, ai_function: AIFunction[_P, _R], *args: _P.args, **kwargs: _P.kwargs) -> dict[str, Any]:
        bound_arguments = ai_function.signature.bind(*args, **kwargs)
        bound_arguments.apply_defaults()
        messages = render_prompt(ai_function.prompt_template, bound_arguments.arguments)
        generation_params = ai_function.generation_params
        model = self._select_model(ai_function.preferred_models)

        tools = list(ai_function.tools)
        if ai_function.return_schema is not None:
            tools.append(ai_function.return_schema)
            tool_mode = ToolUseMode.FORCE
        else:
            tool_mode = ToolUseMode.AUTO
        tool_params = ToolParams(tools, tool_mode)

        return dict(
            model=model,
            messages=messages,
            tool_params=tool_params,
            generation_params=generation_params,
            allow_plain_text_response=ai_function.return_schema is None,
            max_retries_on_invalid_output=ai_function.max_retries_on_invalid_output,
        )

    def execute(self, ai_function: AIFunction[_P, _R], *args: _P.args, **kwargs: _P.kwargs) -> _R:
        call_kwargs = self._prepare_kwargs(ai_function, *args, **kwargs)
        return call_and_use_tools(**call_kwargs)

    async def execute_async(self, ai_function: AIFunction[_P, _R], *args: _P.args, **kwargs: _P.kwargs) -> _R:
        call_kwargs = self._prepare_kwargs(ai_function, *args, **kwargs)
        return await call_and_use_tools_async(**call_kwargs)


def get_runtime() -> Runtime:
    result = _global_runtime.get()
    if result is None:
        raise ValueError("Wanga Runtime is not initialized.")
    return result


_global_runtime: ContextVar[Runtime | None] = ContextVar("global_runtime", default=None)



import collections
import collections.abc
import inspect
import platform
import typing
from dataclasses import dataclass
from datetime import datetime, timedelta
from textwrap import dedent

from attrs import frozen
from pydantic import BaseModel

from wanga.schema.extractor import DEFAULT_SCHEMA_EXTRACTOR
from wanga.schema.jsonschema import JsonSchemaFlavor
from wanga.schema.normalize import normalise_aliases, normalize_annotation, unpack_optional  # noqa: F401
from wanga.schema.schema import (
    CallableSchema,
    LiteralNode,
    ObjectField,
    ObjectNode,
    PrimitiveNode,
    SequenceNode,
    UndefinedNode,
    UnionNode,
)
from wanga.schema.utils import strip_self


def test_normalize_schema():
    expected = {
        list: list,
        list[typing.Annotated[int, "tag"]]: list[int],
        list[typing.List[int]]: list[list[int]],
        list[typing.Tuple[int, str]]: list[tuple[int, str]],
        typing.Union[int, float]: int | float,
        typing.Optional[int]: int | None,
        typing.List: list,
        typing.Union[typing.Union[int, float], str]: int | float | str,
        (typing.Literal[1] | typing.Literal[2] | typing.Literal[3]): typing.Literal[1, 2, 3],
        (typing.Literal[1, 2] | typing.Union[typing.Literal[2, 3], typing.Literal[3, 4]]): (
            typing.Literal[1, 2, 3, 4]
        ),
    }
    for annotation, result in expected.items():
        assert normalize_annotation(annotation) == result


def test_unpack_optional():
    expected = {
        typing.Optional[int]: int,
        int | float | None: int | float,
        int: None,
    }
    for annotation, result in expected.items():
        assert unpack_optional(annotation) == result


def test_concretize_schema():
    expected = {
        typing.List[collections.abc.Iterable[int]]: list[list[int]],
        typing.List[int]: list[int],
        int: int,
        collections.abc.Mapping[str, typing.List[int]]: dict[str, list[int]],
    }
    for annotation, result in expected.items():
        assert normalize_annotation(annotation, concretize=True) == result


def test_extract_schema():
    def foo(x: int, y: str = "hello", z: tuple[int, ...] = ()):  # noqa
        pass

    foo_schema = CallableSchema(
        return_schema=UndefinedNode(original_annotation=typing.Any),
        call_schema=ObjectNode(
            constructor_fn=foo,
            constructor_signature=inspect.signature(foo),
            name="foo",
            hint=None,
            fields=[
                ObjectField(
                    name="x",
                    schema=PrimitiveNode(primitive_type=int),
                    required=True,
                    hint=None,
                ),
                ObjectField(
                    name="y",
                    schema=PrimitiveNode(primitive_type=str),
                    required=False,
                    hint=None,
                ),
                ObjectField(
                    name="z",
                    schema=SequenceNode(
                        sequence_type=tuple,
                        item_schema=PrimitiveNode(primitive_type=int),
                    ),
                    required=False,
                    hint=None,
                ),
            ],
        ),
        long_description=None,
    )

    assert DEFAULT_SCHEMA_EXTRACTOR.extract_schema(foo) == foo_schema

    def bar(x: typing.List[int], y: typing.Literal["hehe"] | float) -> int:  # noqa
        r"""Bar.

        Blah blah blah.

        Args:
            x: The x.
            y: Hard example.
        """
        return 0

    bar_schema = CallableSchema(
        return_schema=PrimitiveNode(primitive_type=int),
        call_schema=ObjectNode(
            constructor_fn=bar,
            constructor_signature=inspect.signature(bar),
            name="bar",
            hint="Bar.",
            fields=[
                ObjectField(
                    name="x",
                    schema=SequenceNode(
                        sequence_type=list,
                        item_schema=PrimitiveNode(primitive_type=int),
                    ),
                    required=True,
                    hint="The x.",
                ),
                ObjectField(
                    name="y",
                    schema=UnionNode(
                        [
                            PrimitiveNode(primitive_type=float),
                            LiteralNode(options=["hehe"]),
                        ]
                    ),
                    required=True,
                    hint="Hard example.",
                ),
            ],
        ),
        long_description="Blah blah blah.",
    )

    assert DEFAULT_SCHEMA_EXTRACTOR.extract_schema(bar) == bar_schema

    class Baz:
        r"""I am Baz."""

        def __init__(self, x: int, y):
            r"""Init Baz.

            Args:
                x: The x.
                y: The y.
            """

    baz_schema = CallableSchema(
        return_schema=UndefinedNode(original_annotation=typing.Any),
        call_schema=ObjectNode(
            constructor_fn=Baz,
            constructor_signature=inspect.signature(Baz),
            name="Baz",
            hint="I am Baz.",
            fields=[
                ObjectField(
                    name="x",
                    schema=PrimitiveNode(primitive_type=int),
                    required=True,
                    hint="The x.",
                ),
                ObjectField(
                    name="y",
                    schema=UndefinedNode(original_annotation=typing.Any),
                    required=True,
                    hint="The y.",
                ),
            ],
        ),
        long_description=None,
    )

    assert DEFAULT_SCHEMA_EXTRACTOR.extract_schema(Baz) == baz_schema

    @frozen
    class Qux:
        r"""I am Qux.

        I have attributes instead of arguments!

        Attributes:
            x: The x.
            baz: The baz.
        """

        x: "int"
        baz: Baz

    qux_schema = CallableSchema(
        return_schema=UndefinedNode(original_annotation=None),
        call_schema=ObjectNode(
            constructor_fn=Qux,
            constructor_signature=inspect.signature(Qux, eval_str=True),
            name="Qux",
            hint="I am Qux.",
            fields=[
                ObjectField(
                    name="x",
                    schema=PrimitiveNode(primitive_type=int),
                    required=True,
                    hint="The x.",
                ),
                ObjectField(
                    name="baz",
                    schema=baz_schema.call_schema,
                    required=True,
                    hint="The baz.",
                ),
            ],
        ),
        long_description="I have attributes instead of arguments!",
    )

    assert DEFAULT_SCHEMA_EXTRACTOR.extract_schema(Qux) == qux_schema

    @dataclass
    class Goo:
        r"""I am Goo.

        I am a dataclass, and I use the stupid ReST docstring syntax!

        :param date: Ohoho, good luck.
        """

        date: datetime

    goo_schema = CallableSchema(
        return_schema=UndefinedNode(original_annotation=None),
        call_schema=ObjectNode(
            constructor_fn=Goo,
            constructor_signature=inspect.signature(Goo),
            name="Goo",
            hint="I am Goo.",
            fields=[
                ObjectField(
                    name="date",
                    schema=ObjectNode(
                        constructor_fn=datetime,
                        constructor_signature=strip_self(inspect.signature(datetime.__init__)),
                        name="datetime",
                        hint=None,
                        fields=[
                            ObjectField(
                                name="year",
                                schema=PrimitiveNode(primitive_type=int),
                                required=True,
                                hint=None,
                            ),
                            ObjectField(
                                name="month",
                                schema=PrimitiveNode(primitive_type=int),
                                required=True,
                                hint=None,
                            ),
                            ObjectField(
                                name="day",
                                schema=PrimitiveNode(primitive_type=int),
                                required=True,
                                hint=None,
                            ),
                            ObjectField(
                                name="hour",
                                schema=PrimitiveNode(primitive_type=int),
                                required=False,
                                hint=None,
                            ),
                            ObjectField(
                                name="minute",
                                schema=PrimitiveNode(primitive_type=int),
                                required=False,
                                hint=None,
                            ),
                            ObjectField(
                                name="second",
                                schema=PrimitiveNode(primitive_type=int),
                                required=False,
                                hint=None,
                            ),
                        ],
                    ),
                    required=True,
                    hint="Ohoho, good luck.",
                ),
            ],
        ),
        long_description="I am a dataclass, and I use the stupid ReST docstring syntax!",
    )

    assert DEFAULT_SCHEMA_EXTRACTOR.extract_schema(Goo) == goo_schema

    class Hoo(BaseModel):
        r"""I am Hoo.

        I am a Pydantic model!
        And I use Numpy Doc format!

        Parameters
        ----------
        delta : timedelta
            Thou shall not pass!
        """

        delta: timedelta

    hoo_schema = CallableSchema(
        return_schema=UndefinedNode(original_annotation=None),
        call_schema=ObjectNode(
            constructor_fn=Hoo,
            constructor_signature=inspect.signature(Hoo),
            name="Hoo",
            hint="I am Hoo.",
            fields=[
                ObjectField(
                    name="delta",
                    schema=ObjectNode(
                        constructor_fn=timedelta,
                        constructor_signature=strip_self(inspect.signature(timedelta.__init__)),
                        name="timedelta",
                        hint=None,
                        fields=[
                            ObjectField(
                                name="days",
                                schema=PrimitiveNode(primitive_type=int),
                                required=False,
                                hint=None,
                            ),
                            ObjectField(
                                name="seconds",
                                schema=PrimitiveNode(primitive_type=int),
                                required=False,
                                hint=None,
                            ),
                        ],
                    ),
                    required=True,
                    hint="Thou shall not pass!",
                ),
            ],
        ),
        long_description="I am a Pydantic model!\nAnd I use Numpy Doc format!",
    )

    assert DEFAULT_SCHEMA_EXTRACTOR.extract_schema(Hoo) == hoo_schema


def test_type_statement():
    _, minor, _ = platform.python_version_tuple()
    if int(minor) < 12:
        return

    expr = r"""
    type A = int | float

    assert normalise_aliases(A) == int | float  # type: ignore

    def foo() -> A:  # type: ignore
        pass

    expected = UnionNode([PrimitiveNode(primitive_type=int), PrimitiveNode(primitive_type=float)])
    assert DEFAULT_SCHEMA_EXTRACTOR.extract_schema(foo).return_schema == expected

    def bar() -> list[A]:  # type: ignore
        pass

    expected = SequenceNode(
        sequence_type=list,
        item_schema=UnionNode([PrimitiveNode(primitive_type=int), PrimitiveNode(primitive_type=float)]),
    )
    assert DEFAULT_SCHEMA_EXTRACTOR.extract_schema(bar).return_schema == expected
    """

    expr = dedent(expr)
    exec(expr)


def test_json_schema():
    @frozen
    class Inner:
        """Inner.

        Long description of Inner.

        Attributes:
            x: The x.
        """

        x: int

    def foo(
        a: int,
        b: str,
        c: Inner,
        d: tuple[int, ...] = (),
        e: typing.Literal["x", "y"] = "x",
        f: str | int = 1,
    ):
        r"""Foo!

        Long description of foo.

        Args:
            a: The a.
            b: The b.
            c: The c.
        """

    expected_json_schema = {
        "name": "foo",
        "description": "Foo!\n\nLong description of foo.",
        "parameters": {
            "type": "object",
            "properties": {
                "a": {"type": "integer", "description": "The a."},
                "b": {"type": "string", "description": "The b."},
                "c": {
                    "type": "object",
                    "properties": {"x": {"type": "integer", "description": "The x."}},
                    "required": ["x"],
                    "description": "The c.\n\nInner.",
                },
                "d": {
                    "type": "array",
                    "items": {"type": "integer"},
                },
                "e": {
                    "type": "string",
                    "enum": ["x", "y"],
                },
                "f": {
                    "type": ["string", "integer"],
                },
            },
            "required": ["a", "b", "c"],
        },
    }

    core_schema = DEFAULT_SCHEMA_EXTRACTOR.extract_schema(foo)
    json_schema = core_schema.json_schema(JsonSchemaFlavor.OPENAI, include_long_description=True)
    assert json_schema == expected_json_schema


def test_eval():
    @frozen
    class Huhu:
        meee: int

    @frozen
    class Hehe:
        hehehe: int
        hohoho: str
        huhuhu: Huhu

    def foo(
        x: float,
        /,
        y: int = 3,
        *,
        z: typing.Literal["a", "b"],
        hehe: Hehe | None = None,
    ):
        assert hehe is not None
        return hehe.huhuhu.meee

    json_input = {
        "x": 1,
        "y": 2,
        "z": "a",
        "hehe": {"hehehe": 3, "hohoho": "haha", "huhuhu": {"meee": 4}},
    }

    core_schema = DEFAULT_SCHEMA_EXTRACTOR.extract_schema(foo)
    assert core_schema.eval(json_input) == 4
